#!/usr/bin/env python
#   encoding: utf-8

# Copyright (C) 2025 D E Haynes
# This file is part of spiki.

# Spiki is free software: you can redistribute it and/or modify it under the terms of the
# GNU General Public License as published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# Spiki is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with spiki.
# If not, see <https://www.gnu.org/licenses/>.

from collections.abc import Callable
from collections.abc import Generator
import contextlib
import copy
import dataclasses
import datetime
import decimal
import logging
from numbers import Number
import os.path
from pathlib import Path
import pkgutil
import shutil
import string
import tempfile
import tomllib
import warnings

from spiki.plugin import Change
from spiki.plugin import Phase


class Visitor(contextlib.ExitStack):

    @staticmethod
    def location_of(node: dict) -> Path:
        try:
            return node["registry"]["index"]["registry"]["path"].resolve()
        except (AttributeError, KeyError, TypeError):
            return node["registry"]["path"].resolve()

    @staticmethod
    def url_of(node: dict) -> str:
        root = node["registry"]["root"]
        parent = Visitor.location_of(node).relative_to(root).parent
        return parent.joinpath(node["metadata"]["slug"]).with_suffix(".html").as_posix()

    def __init__(self, *plugin_types: tuple[Callable], **kwargs):
        super().__init__()
        self.index_name = "index.toml"
        self.state = dict()
        self.running = None
        self.space = None
        self.logger = logging.getLogger("visitor")
        self.plugins = list(filter(None, (self.init_plugin(i) for i in plugin_types)))
        self.options = kwargs

    def __enter__(self):
        self.space = Path(tempfile.mkdtemp()).resolve()
        self.running = [self.enter_context(p) for p in self.plugins]
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        rv = super().__exit__(exc_type, exc_val, exc_tb)
        shutil.rmtree(self.space, ignore_errors=True)
        return rv

    @property
    def root(self) -> Path:
        paths = [i.resolve() for i in self.options.get("paths", [])]
        return Path(os.path.commonprefix(paths))

    def init_plugin(self, type_name: str):
        try:
            cls = pkgutil.resolve_name(type_name)
        except (AttributeError, ModuleNotFoundError) as error:
            self.logger.warning(f"'{type_name}' not resolved. Plugin not loaded.")
            return None

        plugin = cls(self)
        return plugin

    def ancestors(self, path: Path) -> list[Path]:
        return sorted(
            (p for p in self.nodes
             if path.is_relative_to(p.parent)
             and p.name == self.index_name
            ),
            key=lambda x: len(format(x))
        )

    def walk(self, *paths: list[Path]) -> Generator[tuple[Path, dict, str]]:
        paths = [i.resolve() for i in paths]
        for phase in [Phase.CONFIG, Phase.SURVEY]:
            for path in paths:
                for plugin in self.running:
                    try:
                        changes = list(plugin(phase, path=path) or [])
                    except Exception as error:
                        self.logger.warning(error, extra=dict(phase=phase), exc_info=True)
                        continue

                    for change in changes:
                        try:
                            yield dataclasses.replace(change, phase=phase)
                            self.state[change.path] = dataclasses.replace(
                                self.state.setdefault(change.path, change),
                                phase=phase,
                            )
                        except TypeError:
                            continue
            else:
                for change in filter(None, (plugin(phase) for plugin in self.running)):
                    yield dataclasses.replace(change, phase=phase)

        for phase in list(Phase)[2:]:
            for path in list(self.state):
                for plugin in self.running:
                    state = self.state[path]
                    try:
                        change = plugin(phase, path=path, text=state.text, node=state.node, doc=state.doc)
                    except Exception as error:
                        self.logger.warning(error, extra=dict(phase=phase), exc_info=True)
                        continue

                    try:
                        yield dataclasses.replace(change, phase=phase)
                    except TypeError:
                        continue

                    if change.text:
                        self.state[path].text = change.text
                    if change.node:
                        self.state[path].node.update(change.node)
                    if change.doc:
                        self.state[path].doc = change.doc
                    if change.result:
                        self.state[path].result = change.result
            else:
                for change in filter(None, (plugin(phase) for plugin in self.running)):
                    if change.text:
                        self.state.setdefault(change.path, change).text = change.text
                    yield dataclasses.replace(change, phase=phase)



#!/usr/bin/env python
#   encoding: utf-8

# Copyright (C) 2025 D E Haynes
# This file is part of spiki.

# Spiki is free software: you can redistribute it and/or modify it under the terms of the
# GNU General Public License as published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# Spiki is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with spiki.
# If not, see <https://www.gnu.org/licenses/>.

import importlib.resources
import pathlib
import tempfile
import textwrap
import tomllib
import unittest

import spiki
from spiki.plugin import Phase
from spiki.renderer import Renderer
from spiki.visitor import Visitor


class VisitorTests(unittest.TestCase):

    def test_example_atomic(self):
        plugin_types = [
            "spiki.plugins.finder:Finder",
            "spiki.plugins.loader:Loader",
            # "spiki.plugins.indexer:Indexer",
            "spiki.plugins.writer:Writer",
        ]
        examples = importlib.resources.files("spiki.examples")
        witness = []
        with (
            tempfile.TemporaryDirectory() as output_name,
            Visitor(*plugin_types) as visitor,
        ):
            visitor.options = dict(
                output=pathlib.Path(output_name).resolve(),
                paths=[examples.joinpath("atomic")],
            )

            self.assertFalse(visitor.state)
            for change in visitor.walk(*visitor.options["paths"]):
                witness.append(change)

            output_path = pathlib.Path(output_name)
            files = list(output_path.glob("*.html"))

        self.assertEqual(len(visitor.state), 1, visitor.state)
        path = list(visitor.state)[0]
        self.assertEqual("a.toml", path.name)
        self.assertEqual(
            visitor.state[path].node["doc"]["html"]["body"]["main"]["blocks"],
            ["    Hello, World!\n\n    "]
        )

        doc = visitor.state[path].doc
        self.assertIsInstance(doc, str)
        self.assertLess(doc.index("<head"), doc.index("<body"))
        self.assertEqual(doc.count("<meta"), 3)

        self.assertEqual(len([i for i in witness if i.phase == Phase.SURVEY]), 1)
        self.assertEqual(len([i for i in witness if i.phase == Phase.INGEST]), 2)
        self.assertEqual(len([i for i in witness if i.phase == Phase.ENRICH]), 1)
        self.assertEqual(len([i for i in witness if i.phase == Phase.RENDER]), 1)
        self.assertEqual(len([i for i in witness if i.phase == Phase.EXPORT]), 2)

        self.assertEqual(len(files), 1)
        self.assertEqual(files[0].name, "a.html")

    def test_example_basic(self):
        plugin_types = [
            "spiki.plugins.finder:Finder",
            "spiki.plugins.loader:Loader",
            # "spiki.plugins.indexer:Indexer",
            "spiki.plugins.writer:Writer",
        ]
        examples = importlib.resources.files("spiki.examples")
        witness = []
        with (
            tempfile.TemporaryDirectory() as output_name,
            Visitor(*plugin_types) as visitor,
        ):
            visitor.options = dict(
                output=pathlib.Path(output_name).resolve(),
                paths=[examples.joinpath("basic")],
            )

            self.assertFalse(visitor.state)
            for change in visitor.walk(*visitor.options["paths"]):
                witness.append(change)

            output_path = pathlib.Path(output_name)
            files = list(output_path.glob("*.css")) + list(output_path.glob("*.html"))

        self.assertEqual(len(visitor.state), 10, visitor.state)
        path = list(visitor.state)[0]
        self.assertEqual("a.toml", path.name)

        doc = visitor.state[path].doc
        self.assertIsInstance(doc, str)
        self.assertLess(doc.index("<head"), doc.index("<body"))
        self.assertEqual(doc.count("<meta"), 3)
        self.assertLess(doc.index("<nav"), doc.index("<blockquote"), visitor.state[path].node)

        self.assertEqual(len([i for i in witness if i.phase == Phase.SURVEY]), 10)
        self.assertEqual(len([i for i in witness if i.phase == Phase.INGEST]), 20)
        self.assertEqual(len([i for i in witness if i.phase == Phase.ENRICH]), 10)
        self.assertEqual(len([i for i in witness if i.phase == Phase.RENDER]), 10)
        self.assertEqual(len([i for i in witness if i.phase == Phase.EXPORT]), 11)

        file_names = sorted([i.name for i in files])
        self.assertEqual(len(files), 7, files)
        self.assertEqual(file_names[0], "a.html")
        self.assertEqual(file_names[2], "basics.css")



#!/usr/bin/env python
#   encoding: utf-8

# Copyright (C) 2025 D E Haynes
# This file is part of spiki.

# Spiki is free software: you can redistribute it and/or modify it under the terms of the
# GNU General Public License as published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# Spiki is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with spiki.
# If not, see <https://www.gnu.org/licenses/>.

import importlib.resources
import pathlib
import tempfile
import textwrap
import tomllib
import unittest

import spiki
from spiki.plugin import Phase
from spiki.plugins.loader import Loader
from spiki.renderer import Renderer


class LoaderTests(unittest.TestCase):

    def test_slug(self):
        text = "ABab234$%^&*-_ "
        rv = Loader.slugify(text)
        self.assertEqual("abab234-_-", rv)

    def test_stack(self):
        parts = tuple()
        slices = Loader.slices(parts)
        self.assertEqual(slices, [(),])

        parts = (1, 2, 3, 4)
        slices = Loader.slices(parts)
        self.assertEqual(slices, [(), (1,), (1, 2), (1, 2, 3), (1, 2, 3, 4)])

    def test_merge_null(self):
        rv = Loader.merge()
        self.assertEqual(rv, {})

    def test_merge_bases(self):
        a = f"""
        [base.one.two]
        n = 0
        [base.two.one]
        n = 0
        """
        b = f"""
        [[base.two.two]]
        n = 1
        [base.two.one]
        n = 1
        [base.one.two]
        n = 1
        """
        c = f"""
        [[doc.two.two]]
        n = 2
        [doc.two.one]
        n = 2
        """
        d = f"""
        [[doc.two.two]]
        n = 3
        [doc.two.one]
        n = 3
        """
        # Test base defining defaults for doc
        args = [tomllib.loads(i) for i in (a, b, c)]
        rv = Loader.merge(*args)
        self.assertEqual(list(rv["doc"]), ["one", "two"])
        self.assertEqual(list(rv["doc"]["two"]), ["one", "two"])
        self.assertEqual(rv["doc"]["two"]["one"]["n"], 2)
        self.assertEqual([i["n"] for i in rv["doc"]["two"]["two"]], [1, 2])

        # Test doc overriding previous
        arg = tomllib.loads(d)
        rv = Loader.merge(rv, arg)
        self.assertEqual(list(rv["doc"]), ["two"])
        self.assertEqual(list(rv["doc"]["two"]), ["two", "one"])

    def test_combine(self):
        lhs = dict(a=dict(b=1, c=2), b=[dict(d=3, e=4), dict(f=5, g=6)])
        rhs = dict(a=dict(b=10), b=[dict(d=30, e=40)])
        rv = Loader.combine(lhs, rhs)
        self.assertIs(rv, rhs)
        self.assertEqual(rv["a"]["b"], 10)
        self.assertEqual(rv["a"]["c"], 2)
        self.assertEqual(len(rv["b"]), 3)

    def test_merge_base(self):
        index_toml = textwrap.dedent("""
        [base]
        config = {tag_mode = "pair"}

        [[base.html.body.nav.ul.li]]
        attrib = {href = "/"}
        a = "Home"

        """)
        index = tomllib.loads(index_toml)

        node_toml = textwrap.dedent("""
        [[doc.html.body.nav.ul.li]]
        attrib = {href = "/faq.html"}
        a = "FAQ"

        """)
        node = tomllib.loads(node_toml)

        template = Loader.merge(index, node)
        self.assertEqual(template["doc"]["config"]["tag_mode"], "pair")
        rv = Renderer().serialize(template)
        self.assertEqual(rv.count("href"), 2, rv)

    def test_merge_index_content(self):
        index_toml = textwrap.dedent("""
        [base]
        config = {tag_mode = "pair", block_wrap = "div"}

        [[base.html.body.nav.ul.li]]
        attrib = {href = "/"}
        a = "Home"
        """)
        index = tomllib.loads(index_toml)

        node_toml = textwrap.dedent("""

        [doc.html]
        config = {tag_mode = "pair"}

        [[doc.html.body.main.dl.div]]
        dt = "Title"
        dd = "Test"

        [[doc.html.body.main.dl.div]]
        dt = "Version"
        dd = "1"

        [[doc.html.body.nav.ul.li]]
        attrib = {href = "/faq.html"}
        a = "FAQ"

        """)
        node = tomllib.loads(node_toml)

        template = Loader.merge(index, node)
        self.assertEqual(template["doc"]["config"]["tag_mode"], "pair")
        rv = Renderer().serialize(template)
        self.assertEqual(rv.count("href"), 2, rv)
        self.assertEqual(rv.count("<div"), 2, rv)



#!/usr/bin/env python
#   encoding: utf-8

# Copyright (C) 2025 D E Haynes
# This file is part of spiki.

# Spiki is free software: you can redistribute it and/or modify it under the terms of the
# GNU General Public License as published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# Spiki is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with spiki.
# If not, see <https://www.gnu.org/licenses/>.


import dataclasses
import enum
import logging
from pathlib import Path
import string


class Phase(enum.Enum):
    CONFIG = "Configuring framework"
    SURVEY = "Discovering topology"
    INGEST = "Reading structured data"
    ENRICH = "Attaching metadata"
    EXTEND = "Hierarchical influences"
    FILTER = "Selecting sources"
    ASSETS = "Preparing media"
    ROUTES = "Interconnections"
    EFFECT = "Adaptations and modifications"
    RENDER = "Generating content"
    EXPORT = "Finalizing output"
    REPORT = "Summary"


@dataclasses.dataclass
class Change:
    object: object  = None
    phase:  Phase   = None
    path:   Path    = None
    type:   str     = ""
    text:   str     = None
    node:   dict    = dataclasses.field(default_factory=dict)
    doc:    str     = None
    result: Path    = None


class Plugin:

    def __init__(self, visitor: "Pathfinder" = None):
        self.logger = logging.getLogger(self.__class__.__name__.lower())
        self.visitor = visitor
        self.phase = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False

    def __call__(self, phase: Phase, *, path: Path = None, node: dict = None, doc: str = None, **kwargs) -> Change:
        self.phase = phase
        if path is None:
            method = getattr(self, f"end_{phase.name.lower()}", None)
        elif phase == Phase.SURVEY:
            method = getattr(self, f"gen_{phase.name.lower()}", None)
        else:
            method = getattr(self, f"run_{phase.name.lower()}", None)

        if method:
            rv = method(path=path, node=node, doc=doc, **kwargs)
            return rv or Change(self, phase=phase, path=path, node=node, doc=doc)
        else:
            return None

    @staticmethod
    def slugify(text: str, table="".maketrans({i: i for i in string.ascii_letters + string.digits + "_-"})):
        mapping = {ord(i): None for i in text}
        mapping.update(table)
        mapping[ord(" ")] = "-"
        return text.translate(mapping).lower()



#!/usr/bin/env python
#   encoding: utf-8

# Copyright (C) 2025 D E Haynes
# This file is part of spiki.

# Spiki is free software: you can redistribute it and/or modify it under the terms of the
# GNU General Public License as published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# Spiki is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with spiki.
# If not, see <https://www.gnu.org/licenses/>.

import argparse
from collections import deque
import html
import itertools
import operator
import re
import sys
import textwrap

from . import __version__

__doc__ = """"

From the command line::

    echo "Hello, World!" | python -m speechmark

    <blockquote>
    <p>
    Hello World!
    </p>
    </blockquote>

"""

class SpeechMark:
    """
    Parsing text programmatically::

        from speechmark import SpeechMark

        text = '''
        <PHONE.announcing@GUEST,STAFF> Ring riiing!
        <GUEST:thinks> I wonder if anyone is going to answer that phone.
        '''.strip()

        sm = SpeechMark()
        sm.loads(text)

    ... produces this HTML5 output::

        <blockquote cite="&lt;PHONE.announcing@GUEST,STAFF&gt;">
        <cite data-role="PHONE" data-directives=".announcing@GUEST,STAFF">PHONE</cite>
        <p>
         Ring riiing!
        </p>
        </blockquote>
        <blockquote cite="&lt;GUEST:thinks&gt;">
        <cite data-role="GUEST" data-mode=":thinks">GUEST</cite>
        <p>
         I wonder if anyone is going to answer that phone.
        </p>
        </blockquote>

    """
    def __init__(
        self,
        lines=[],
        maxlen=None,
        noescape="!\"',-;{}~",
    ):
        self.cue_matcher = re.compile(
            r"""
        ^<                              # Opening bracket
        (?P<role>[^\.:\\?# >]*)         # Role
        (?P<directives>[^\:\\?# >]*)    # Directives
        (?P<mode>[^\\?# >]*)            # Mode
        (?P<parameters>[^# >]*)         # Parameters
        (?P<fragments>[^ >]*)           # Fragments
        >                               # Closing bracket
        """,
            re.VERBOSE,
        )

        self.list_matcher = re.compile(
            r"""
        ^\s*                            # Leading space
        (?P<ordinal>\+|\d+\.)           # Digits and a dot
        """,
            re.VERBOSE,
        )

        self.tag_matcher = re.compile(
            r"""
        (?P<tag>[`*_])(?P<text>.*?)(?P=tag) # Non-greedy pair
        """,
            re.VERBOSE,
        )
        self.tagging = {"`": "code", "_": "strong", "*": "em"}

        self.link_matcher = re.compile(
            r"""
        \[(?P<label>[^\]]*?)\]          # Non-greedy, permissive
        \((?P<link>[^\)]*?)\)           # Non-greedy, permissive
        """,
            re.VERBOSE,
        )

        self.escape_table = str.maketrans(
            {
                v: f"&{k}"
                for k, v in html.entities.html5.items()
                if k.endswith(";") and len(v) == 1 and v not in noescape + "#+.`_*[]()@?=:/"
            }
        )
        self.source = deque(lines, maxlen=maxlen)
        self._index = 0

    @property
    def text(self) -> str:
        return "\n".join(self.source)

    def cue_element(self, match: re.Match) -> str:
        details = match.groupdict()
        if not details["role"].strip() and not details["parameters"]:
            return ""

        attrs = " ".join(
            f'data-{k}="{html.escape(v, quote=True)}"' for k, v in details.items() if v.strip()
        )
        return f"<cite{' ' if attrs else ''}{attrs}>{details['role']}</cite>"

    def tag_element(self, match: re.Match) -> str:
        details = match.groupdict()
        tag = self.tagging.get(details.get("tag"), "")
        return f"<{tag}>{details['text'].translate(self.escape_table)}</{tag}>"

    def link_element(self, match: re.Match) -> str:
        details = match.groupdict()
        href = html.escape(details["link"], quote=True)
        return f"""<a href="{href}">{details['label'].translate(self.escape_table)}</a>"""

    def parse_lines(self, terminate: bool):
        # Make a local list of lines to process
        lines = list(itertools.islice(self.source, self._index, None))

        # Match any available cues and store them by line number
        cues = dict(
            filter(
                operator.itemgetter(1),
                ((n, self.cue_matcher.match(line)) for n, line in enumerate(lines)),
            )
        )

        # Create a sequence of 2-tuples which demarcate each block
        blocks = list(
            itertools.pairwise(
                sorted(set(cues.keys()).union({0, len(lines)} if terminate else {0}))
            )
        )

        end = 0
        for begin, end in blocks:
            cue = cues.get(begin)
            text = "\n".join(lines[begin:end]).rstrip()
            yield "\n".join(
                i
                for i in self.parse_block(cue, text.splitlines(keepends=False), terminate)
                if isinstance(i, str)
            )

        self._index = end

    def parse_block(self, cue, lines, terminate=False):
        list_items = dict(
            filter(
                operator.itemgetter(1),
                ((n, self.list_matcher.match(line)) for n, line in enumerate(lines)),
            )
        )

        list_type = ""
        paragraph = False
        for n, line in enumerate(lines):
            if n == 0:
                if cue:
                    yield '<blockquote cite="{0}">'.format(html.escape(cue.group(), quote=True))
                    yield self.cue_element(cue)
                elif not n:
                    yield "<blockquote>"

            if line.lstrip().startswith("#"):
                yield f"<!-- {line.translate(self.escape_table)} -->"
                continue

            if n in list_items:
                item = list_items[n]
                yield item
                details = item.groupdict()
                if list_type:
                    yield "</p></li>"
                else:
                    list_type = "ul" if details["ordinal"].strip() == "+" else "ol"
                    if paragraph:
                        yield "</p>"
                    yield f"<{list_type}>"

                if list_type == "ul":
                    yield f"<li><p>"
                else:
                    yield f"""<li id="{details['ordinal'].rstrip('.')}"><p>"""
                line = line[item.end() :].lstrip()  # Retain hanging text

            elif not paragraph and n < min(list_items or [sys.maxsize]):
                paragraph = True
                yield f"<p>"
            elif not line:
                yield "</p>"
                yield "<p>"

            # Collect all matches and their substitutions
            subs = dict(
                (m.span(), fn(m))
                for fn, i in (
                    (self.cue_element, self.cue_matcher),
                    (self.link_element, self.link_matcher),
                    (self.tag_element, self.tag_matcher),
                )
                for m in i.finditer(line)
            )

            # Create spans for unmatched chunks of text
            chunks = list(
                itertools.pairwise(
                    sorted({pos for span in subs for pos in span} | {0, len(line)})
                )
            )

            # Add default processing for each unmatched span
            for span in chunks:
                if span not in subs:
                    subs[span] = line[span[0] : span[1]].translate(self.escape_table)

            line = "".join("" if cue and cue.span() == span else subs[span] for span in chunks)
            yield line

        if terminate:
            if list_type:
                yield "</p></li>"
                yield f"</{list_type}>"
            elif paragraph:
                yield "</p>"
                paragraph = False
            yield "</blockquote>"

    def loads(self, text: str, marker: str = "\n", **kwargs) -> str:
        self.reset()
        result = marker.join(i.strip() for i in self.feed(text, terminate=True))
        return f"{result}{marker}"

    def feed(self, text: str, terminate=False, **kwargs):
        self.source.extend(text.splitlines(keepends=False))
        yield from self.parse_lines(terminate)

    def reset(self):
        self.source.clear()
        self._index = 0


def parser():
    rv = argparse.ArgumentParser(
        usage="\n".join((__doc__, textwrap.dedent(SpeechMark.__doc__)))
    )

    rv.add_argument(
        "--version", action="store_true", default=False,
        help="display the package version"
    )

    return rv


def main(args):
    if args.version:
        print(__version__)
        return 0

    sm = SpeechMark()
    text = sys.stdin.read()
    html5 = sm.loads(text)
    print(html5)
    return 0


def run():
    p = parser()
    args = p.parse_args()
    rv = main(args)
    sys.exit(rv)


if __name__ == "__main__":
    run()



#!/usr/bin/env python
#   encoding: utf-8

# Copyright (C) 2025 D E Haynes
# This file is part of spiki.

# Spiki is free software: you can redistribute it and/or modify it under the terms of the
# GNU General Public License as published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# Spiki is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with spiki.
# If not, see <https://www.gnu.org/licenses/>.

import html
import re
import textwrap
import tomllib
import unittest

import spiki
from spiki.speechmark import SpeechMark

__doc__ = f"""
:Version: {spiki.__version__}
:Author: D E Haynes
:Licence: `CC BY-NC-ND <https://creativecommons.org/licenses/by-nc-nd/4.0/>`_ Attribution-NonCommercial-NoDerivs

"""

# Ex 2
"""
<blockquote>
<header>Mandelbrot </header>
<p><span class="text">Look, if you can't make a proper analysis of the situation, why don't you&NewLine;pick at </span>
<a href="https://www.random.org/integers/?num=1&min=1&max=3&col=1&base=10&format=html">random</a>
<span class="text">&quest;</span></p>
</blockquote>
"""

# Ordered list marker: '[0-9a-zA-Z].' (remember fragment reference)
# Unordered list marker: '+'
# <em>:  one or more *
# <strong>: _
# <code>:  `
# <a>: [label](url)

Head = ("propose", "confirm", "counter", "abandon", "condemn", "declare")
Hand = ("decline", "suggest", "promise", "disavow", "deliver")


class SupportTests(unittest.TestCase):

    def test_elaboration_directive(self):
        text = """
        <ALAN.elaborating> Maybe now's a good time to ask {BETH.name} a question.
            1. Ask about the weather
            2. Ask about pets
            3. Ask about football
        """
        sm = SpeechMark()
        rv = sm.loads(text)
        self.assertIn("Ask about football", rv)


class Syntax(unittest.TestCase):
    """
    SpeechMark
    ##########

    SpeechMark is a convention for markup of authored text.
    It is designed for capturing dialogue, attributing speech, and writing screenplay directions.
    This document explains the syntax, and shows how it should be rendered in HTML5.

    SpeechMark takes inspiration from other markup systems already in common use, eg:

    * `Markdown <https://commonmark.org/>`_
    * `reStructuredText <https://docutils.sourceforge.io/rst.html>`_

    I tried both these systems prior to creating SpeechMark. I found I needed some features which
    Markdown didn't have. RST proved to be overkill for this particular purpose, and the document model
    became cumbersome to me.

    Philosophy
    ==========

    SpeechMark syntax is deliberately constrained to be simple and unambiguous.
    This is to permit fast and efficient processing of many small pieces of text over an extended period of time.

    SpeechMark does not concern itself with document structure. There are no titles, sections or breaks.
    Rather, the input is expected to be a stream of text fragments.

    The specification intends to be lossless, so that every non-whitespace feature of the original text
    may be retrieved from the output. It should be possible to round-trip your SpeechMark scripts into
    HTML5 and back again.

    Features
    ========

    SpeechMark has the basic elements you see in other markup systems, ie:

        * Emphasis_
        * Hyperlinks_
        * Comments_
        * Lists_

    There is one feature very specific to SpeechMark:

        * Cues_

    SpeechMark doesn't try to do everything. To integrate it into an application, you may
    need:

        * Preprocessing_
        * Postprocessing_

    Emphasis
    --------

    SpeechMark supports three flavours of emphasis.

    * Surround text by asterisks ``*like this*`` to generate ``<em>`` tags.
    * Use underscores ``_like this_`` to generate ``<strong>`` tags.
    * Use backticks ```like this``` to generate ``<code>`` tags.

    Hyperlinks
    ----------

    Hyperlinks have two components; the label and the URL.
    The label appears first within square brackets, followed by the URL in parentheses::

        [SpeechMark](https://github.com/thuswise/spiki)

    Comments
    --------

    The `#` character denotes a comment. It must be the first character on a line::

        # Comments aren't ignored. They get converted to HTML (<!-- -->)

    Lists
    -----

    Unordered lists
    ```````````````

    The `+` character creates a list item of the text which follows it, like so::

        + Beef
        + Lamb
        + Fish


    Ordered lists
    `````````````
    Using digits and a dot before text will give you an ordered list::

        1. Beef
        2. Lamb
        3. Fish

    Cues
    ----

    A cue marks the start of a new block of dialogue. Is is denoted by angled brackets::

        <>  Once upon a time, far far away...

    Cues are flexible structures. They have a number of features you can use all together, or
    you can leave them empty.

    A cue may contain information about the speaker of the dialogue, and how they deliver it.

    The most basic of these is the **role**. This is the named origin of the lines of dialogue.
    It is recommended that you state the role in upper case letters, eg: GUEST, STAFF.
    Inanimate objects can speak too of course. Eg: KETTLE, and PHONE::

        <PHONE> Ring riiing!

    The **mode** declares the form in which the act of speech is delivered.
    Although it's the most common, *says* is just one of many possible modes of speech.
    There are others you might want to use, like *whispers* or *thinks*.
    The mode is separated by a colon::

        <GUEST:thinks> I wonder if anyone is going to answer that phone.

    Capturing the mode of speech enables different presentation options,
    eg: character animations to match the delivery.
    Modes of speech should be stated in the simple present, third person form.

    **Directives** indicate that there are specific side-effects to the delivery of the dialogue.
    They may be used to fire transitions in a state machine, specifying that the speech achieves
    progress according to some social protocol.

    It's recommended that these directives be stated as present participles
    such as *promising* or *declining*::

        <PHONE.announcing> Ring riiing!

    Directives, being transitive in nature, sometimes demand objects to their action. So you may
    specify the recipient roles of the directive if necessary too::

        <PHONE.announcing@GUEST,STAFF> Ring riiing!

    **Parameters** are key-value pairs which modify the presentation of the dialogue. SpeechMark borrows the
    Web URL syntax for parameters (first a '?', with '&' as the delimiter).

    Their meaning is specific to the application. For example, it might be necessary to specify
    some exact timing for the revealing of the text::

        <?pause=3&dwell=0.4>

            Above, there is the sound of footsteps.

            Of snagging on a threadbare carpet.

            Then shuffling down the ancient stairs.

    SpeechMark recognises the concept of **fragments**, which also come from URLs. That's the part after a '#'
    symbol. You can use the fragment to refer to items in a list::

        <STAFF.proposing#3> What will you have, sir? The special is fish today.

            1. Order the Beef Wellington
            2. Go for the Shepherd's Pie
            3. Try the Dover Sole

    Preprocessing
    =============

    Whitespace
    ----------

    A SpeechMark parser expects certain delimiters to appear only at the beginning of a line.
    Therefore, if your marked-up text has been loaded from a file or data structure, you may need to
    remove any common indentation and trim the lines of whitespace characters.

    Variable substitution
    ---------------------

    It would be very handy for dialogue to reference some objects in scope.
    That would allow us to make use of their attributes, eg: ``GUEST.surname``.

    Unfortunately, the syntax for variable substitution is language dependent.
    Equally the mode of attribute access is application dependent.
    Should it be ``GUEST.surname`` or ``GUEST['surname']``?

    SpeechMark therefore does not provide this ability, and it must be performed prior to parsing.
    Here's an example using Python string formatting, where the context variables are dictionaries::

        <GUEST> I'll have the Fish, please.

        <STAFF> Very good, {GUEST['honorific']} {GUEST['surname']}.


    Postprocessing
    ==============

    Pruning
    -------

    SpeechMark tries not to throw anything away. You might not want that behaviour. Specifically,
    you may prefer to remove lines of comment from the HTML5 output.

    Since the output is line-based, it's a simple matter to strip out those lines using your favourite programming
    language or command line tools.

    Extending
    ---------

    SpeechMark does not support extensions. There is no syntax to create custom tags.

    However, if you need to transform the output before it gets to the web, you could utilise the
    ``<code>`` tag for that purpose.

    Suppose you have a menu you've defined as a list::

        + `button`[Map](/api/map)
        + `button`[Inventory](/api/inventory)

    Here is part of that output::

        <li><p><code>button</code><a href="/api/map">Map</a></p></li>

    This could be sufficient to trigger a ``button`` function in your postprocessor which replaces
    the bare link with a ``<form>`` and ``<input>`` controls to pop up the map.

    Specification
    =============

    1. General
    ----------

    1.1
    ```

    SpeechMark input must be line-based text, and should have UTF-8 encoding.

    1.2
    ```

    Inline markup must consist of pairs of matching delimiters. There must be no line break within them;
    all inline markup must terminate on the same line where it begins. Delimiters may not contain other
    delimiter pairs. There is no nested markup.

    1.3
    ```

    The generated output must be one or more HTML5 ``blockquote`` elements.
    All elements must be explicitly terminated.

    1.4
    ```

    All output must be placed within blocks. Each block may begin with a cite element. A block may contain one
    or more paragraphs. A block may contain a list. Every list item must contain a paragraph.

    """

    examples = []

    def example(label=""):
        def wrapper(fn):
            def inner(self, *args, **kwargs):
                return fn(self, **data)

            doc = fn.__doc__ or ""
            text, toml = re.split(r"#\W*TOML\n", doc, maxsplit=1)
            data = tomllib.loads(toml)
            Syntax.examples.append((label, text, data, fn))
            fn.__doc__ = text
            return inner

        return wrapper

    def check(self, markup: dict = {}, output=""):
        sm = SpeechMark()
        for n, (tag, text) in enumerate(markup.items()):
            text = textwrap.dedent(text).strip()
            with self.subTest(n=n, tag=tag, text=text):
                rv = sm.loads(text)
                self.compare(rv, output)

    def compare(self, a, b, msg=""):
        x = "".join(i.strip() for i in a.splitlines())
        y = "".join(i.strip() for i in b.splitlines())
        self.assertEqual(x, y, msg or a)


class EmphasisTests(Syntax):
    """
    2. Emphasis
    -----------

    """

    @Syntax.example(label="2.01")
    def test_minimal_emphasis(self, markup: dict = {}, output=""):
        """
        Emphasis is added using pairs of asterisks.

        # TOML
        markup."Single instance"  =   "*Definitely!*"
        output = '''
        <blockquote>
        <p><em>Definitely!</em></p>
        </blockquote>
        '''
        """
        return self.check(markup, output)

    @Syntax.example(label="2.02")
    def test_multiple_emphasis(self, markup: dict = {}, output=""):
        """
        There may be multiple emphasized phrases on a line.

        # TOML
        markup."Multiple instances" =   "*Definitely* *Definitely!*"
        output = '''
        <blockquote>
        <p><em>Definitely</em> <em>Definitely!</em></p>
        </blockquote>
        '''
        """
        return self.check(markup, output)

    def test_cornercases_abutting_emphasis(self):
        expected = textwrap.dedent(
            """
        <blockquote>
        <p><em>Definitely</em><em>Definitely!</em></p>
        </blockquote>
        """
        )
        sm = SpeechMark()
        rv = sm.loads("*Definitely**Definitely!*")
        self.compare(rv, expected, rv)


class SignificanceTests(Syntax):
    @Syntax.example(label="2.03")
    def test_minimal_significance(self, markup: dict = {}, output=""):
        """
        Strong text is denoted with underscores.

        # TOML
        markup."Single instance"  =   "_Warning!_"
        output = '''
        <blockquote>
        <p><strong>Warning!</strong></p>
        </blockquote>
        '''
        """
        return self.check(markup, output)

    @Syntax.example(label="2.04")
    def test_multiple_significance(self, markup: dict = {}, output=""):
        """
        There may be multiple snippets of significant text on one line.

        # TOML
        markup."Multiple instances" =   "_Warning_ _Warning_!"
        output = '''
        <blockquote>
        <p><strong>Warning</strong> <strong>Warning</strong>!</p>
        </blockquote>
        '''
        """
        return self.check(markup, output)

    def test_cornercases_code(self):
        expected = textwrap.dedent(
            """
        <blockquote>
        <p>
        <strong>Warning</strong><strong>Warning</strong>!
        </p>
        </blockquote>
        """
        )
        sm = SpeechMark()
        rv = sm.loads("_Warning__Warning_!")
        self.compare(rv, expected, rv)


class CodeTests(Syntax):
    @Syntax.example(label="2.05")
    def test_single_code(self, markup: dict = {}, output=""):
        """
        Code snippets are defined between backticks.

        # TOML
        markup."Single instance"  =   "`git log`"
        output = '''
        <blockquote>
        <p><code>git log</code></p>
        </blockquote>
        '''
        """
        return self.check(markup, output)

    @Syntax.example(label="2.06")
    def test_multiple_code(self, markup: dict = {}, output=""):
        """
        There may be multiple code snippets on a line.

        # TOML
        markup."Multiple instances" =   "`git` `log`"
        output = '''
        <blockquote>
        <p><code>git</code> <code>log</code></p>
        </blockquote>
        '''
        """
        return self.check(markup, output)

    def test_cornercases_code(self):
        expected = textwrap.dedent(
            """
        <blockquote>
        <p><code>8.8.8.8</code></p>
        </blockquote>
        """
        )
        sm = SpeechMark()
        rv = sm.loads("`8.8.8.8`")
        self.compare(rv, expected, rv)

    def test_cornercases_abutted_code(self):
        expected = textwrap.dedent(
            """
        <blockquote>
        <p><code>git</code><code>log</code></p>
        </blockquote>
        """
        )
        sm = SpeechMark()
        rv = sm.loads("`git``log`")
        self.compare(rv, expected, rv)


class LinkTests(Syntax):
    """
    3. Hyperlinks
    -------------

    """

    def test_basic_match(self):
        sm = SpeechMark()
        match = sm.link_matcher.match("[Python](https://python.org)")
        self.assertTrue(match)
        self.assertEqual("Python", match.groupdict().get("label"))
        self.assertEqual("https://python.org", match.groupdict().get("link"))

    @Syntax.example(label="3.01")
    def test_single_link(self, markup: dict = {}, output=""):
        """
        Hyperlinks are defined by placing link text within square brackets and the link destination
        in parentheses. There must be no space between them.
        See also https://spec.commonmark.org/0.30/#example-482.

        # TOML
        markup."Single instance"  =    "[Python](https://python.org)"
        output = '''
        <blockquote>
        <p><a href="https://python.org">Python</a></p>
        </blockquote>
        '''
        """
        return self.check(markup, output)

    @Syntax.example(label="3.02")
    def test_multiple_links(self, markup: dict = {}, output=""):
        """
        There may be multiple hyperlinks on a line.

        # TOML
        markup."Multiple instances" =  "[Python](https://python.org) [PyPI](https://pypi.org)"
        output = '''
        <blockquote>
        <p><a href="https://python.org">Python</a> <a href="https://pypi.org">PyPI</a></p>
        </blockquote>
        '''
        """
        return self.check(markup, output)

    def test_cornercases_links_with_spaces(self):
        expected = textwrap.dedent(
            """
        <blockquote>
        <p>[Python] (https://python.org)</p>
        </blockquote>
        """
        )
        sm = SpeechMark()
        rv = sm.loads("[Python] (https://python.org)")
        self.compare(rv, expected, rv)

    def test_cornercases_abutting_links(self):
        expected = textwrap.dedent(
            """
        <blockquote>
        <p><a href="https://python.org">Python</a><a href="https://python.org">Python</a></p>
        </blockquote>
        """
        )
        sm = SpeechMark()
        rv = sm.loads("[Python](https://python.org)[Python](https://python.org)")
        self.compare(rv, expected, rv)


class CommentTests(Syntax):
    """
    4. Comments
    -----------

    """

    @Syntax.example(label="4.01")
    def test_single_comment(self, markup: dict = {}, output=""):
        """
        Any line beginning with a "#" is a comment.
        It is output in its entirety (including delimiter) as an HTML comment.

        # TOML
        markup."Single instance"  =    "# TODO"
        output = '''
        <blockquote>
        <!-- # TODO -->
        </blockquote>
        '''
        """
        return self.check(markup, output)


class UnorderedListTests(Syntax):
    """
    5. Lists
    --------

    """

    @Syntax.example(label="5.01")
    def test_minimal_list(self, markup: dict = {}, output=""):
        """
        A line beginning with a '+' character constitutes an
        item in an unordered list.

        # TOML
        markup."Single list"  =  '''
        + Hat
        + Gloves
        '''
        output = '''
        <blockquote>
        <ul>
        <li><p>Hat</p></li>
        <li><p>Gloves</p></li>
        </ul>
        </blockquote>
        '''
        """
        return self.check(markup, output)


class OrderedListTests(Syntax):
    @Syntax.example(label="5.02")
    def test_numbered_list(self, markup: dict = {}, output=""):
        """
        Ordered lists have lines which begin with one or more digits. Then a dot, and at least one space.

        # TOML
        markup."Single list"  =  '''
        1. Hat
        2. Gloves
        '''
        output = '''
        <blockquote>
        <ol>
        <li id="1"><p>Hat</p></li>
        <li id="2"><p>Gloves</p></li>
        </ol>
        </blockquote>
        '''
        """
        return self.check(markup, output)

    @Syntax.example(label="5.03")
    def test_zeropadded_list(self, markup: dict = {}, output=""):
        """
        Ordered list numbering is exactly as declared. No normalization is performed.

        # TOML
        markup."Single list"  =  '''
        01. Hat
        02. Gloves
        '''
        output = '''
        <blockquote>
        <ol>
        <li id="01"><p>Hat</p></li>
        <li id="02"><p>Gloves</p></li>
        </ol>
        </blockquote>
        '''
        """
        return self.check(markup, output)

    def test_list_matching_positive(self):
        examples = [
            "+Hat",
            "+ Hat",
            "+ <Hat>",
            "1.Hat",
            "1. Hat",
            "1. <Hat>",
            "01.Hat",
            "01. Hat",
            "01. <Hat>",
        ]
        sm = SpeechMark()
        for line in examples:
            with self.subTest(line=line):
                rv = sm.list_matcher.match(line)
                self.assertTrue(rv)
                d = rv.groupdict()
                self.assertEqual(1, len(d))

    def test_list_matching_negative(self):
        examples = [
            "<> +",
        ]
        sm = SpeechMark()
        for line in examples:
            with self.subTest(line=line):
                rv = sm.list_matcher.match(line)
                self.assertFalse(rv)


class CueTests(Syntax):
    """
    6. Cues
    -------

    A cue mark generates a new block.

    6.01
    ````

    A cue mark must appear at the start of a line. No whitespace is allowed in a cue mark.
    A generated ``blockquote`` tag may store the original cue string in its ``cite`` attribute.
    The string must be appropriately escaped.

    """

    @Syntax.example(label="6.02")
    def test_empty_cue(self, markup: dict = {}, output=""):
        """
        All components of a cue are optional.

        # TOML
        markup."Anonymous cue"  =    "<> Once upon a time, far, far away..."
        output = '''
        <blockquote cite="&lt;&gt;">
        <p>Once upon a time, far, far away...</p>
        </blockquote>
        '''
        """
        return self.check(markup, output)

    @Syntax.example(label="6.03")
    def test_role_only_cue(self, markup: dict = {}, output=""):
        """
        It is recommended that roles be stated in upper case.
        When a role is stated, a ``cite`` element must be generated.
        The value of the role must be stored in the ``data-role`` attribute of the cite tag.
        The role value must be appropriately escaped.

        # TOML
        markup."Role only"  =    "<PHONE> Ring riiing!"
        output = '''
        <blockquote cite="&lt;PHONE&gt;">
        <cite data-role="PHONE">PHONE</cite>
        <p>Ring riiing!</p>
        </blockquote>
        '''
        """
        return self.check(markup, output)

    @Syntax.example(label="6.04")
    def test_role_with_mode(self, markup: dict = {}, output=""):
        """
        A mode is preceded by a colon. It is stated after any role.
        When a mode is stated, a ``cite`` element must be generated.
        The value of the mode must be stored in the ``data-mode`` attribute of the cite tag.
        The mode value retains its delimiter. The mode value must be appropriately escaped.
        Modes of speech should be stated in the third person simple present form.

        # TOML
        markup."Role with mode" =   "<GUEST:thinks> I wonder if anyone is going to answer that phone."
        output = '''
        <blockquote cite="&lt;GUEST:thinks&gt;">
        <cite data-role="GUEST" data-mode=":thinks">GUEST</cite>
        <p>I wonder if anyone is going to answer that phone.</p>
        </blockquote>
        '''
        """
        return self.check(markup, output)

    @Syntax.example(label="6.05")
    def test_role_with_directive(self, markup: dict = {}, output=""):
        """
        There may be multiple directives, each preceded by a dot. They are stated after any role.
        When a directive is stated, a ``cite`` element must be generated.
        The directives must be stored in the ``data-directives`` attribute of the cite tag.
        They retain their delimiters. The directives value must be appropriately escaped.
        Directives should be stated as present participles.

        # TOML
        markup."Role with directive" =  "<PHONE.announcing> Ring riiing!"
        output = '''
        <blockquote cite="&lt;PHONE.announcing&gt;">
        <cite data-role="PHONE" data-directives=".announcing">PHONE</cite>
        <p>Ring riiing!</p>
        </blockquote>
        '''
        """
        return self.check(markup, output)

    @Syntax.example(label="6.06")
    def test_role_with_recipients(self, markup: dict = {}, output=""):
        """
        When a directive is stated, a recipient list may follow it. A recipient list begins with a ``@`` symbol.
        The items in the list are separated by commas.
        The recipients must be stored in the ``data-directives`` attribute of the cite tag.
        They retain their delimiters. The directives value must be appropriately escaped.
        Recipients should be stated elsewhere as roles.

        # TOML
        markup."Role with directive and recipients" =  "<PHONE.announcing@GUEST,STAFF> Ring riiing!"
        output = '''
        <blockquote cite="&lt;PHONE.announcing@GUEST,STAFF&gt;">
        <cite data-role="PHONE" data-directives=".announcing@GUEST,STAFF">PHONE</cite>
        <p>Ring riiing!</p>
        </blockquote>
        '''
        """
        return self.check(markup, output)

    @Syntax.example(label="6.07")
    def test_parameters_only(self, markup: dict = {}, output=""):
        """
        A parameter list begins with a ``?`` symbol. It consists of ``key=value`` pairs separated by ampersands.
        Should a directive be stated, any parameter list must come after it.
        The parameters must be stored in the ``data-parameters`` attribute of the cite tag.
        They retain their delimiters. The parameters value must be appropriately escaped.

        # TOML
        markup."Parameters only" =  "<?pause=3&dwell=0.4> Above, there is the sound of footsteps."
        output = '''
        <blockquote cite="&lt;?pause=3&amp;dwell=0.4&gt;">
        <cite data-parameters="?pause=3&amp;dwell=0.4"></cite>
        <p>Above, there is the sound of footsteps.</p>
        </blockquote>
        '''
        """
        return self.check(markup, output)

    @Syntax.example(label="6.08")
    def test_directive_and_fragment(self, markup: dict = {}, output=""):
        """
        There may be multiple fragments. The first begins with a ``#`` symbol.
        All semantics are those of `Web URLs <https://url.spec.whatwg.org>`_.
        The fragments appear at the end of any cue mark.
        The fragments must be stored in the ``data-fragments`` attribute of the cite tag.
        They retain all delimiters. The fragments value must be appropriately escaped.

        # TOML
        markup."Role with directive and fragment" =  '''
        <STAFF.proposing#3> What will you have, sir? The special is fish today.
            1. Order the Beef Wellington
            2. Go for the Shepherd's Pie
            3. Try the Dover Sole
        '''
        output = '''
        <blockquote cite="&lt;STAFF.proposing#3&gt;">
        <cite data-role="STAFF" data-directives=".proposing" data-fragments="#3">STAFF</cite>
        <p>What will you have, sir? The special is fish today.</p>
        <ol>
        <li id="1"><p>Order the Beef Wellington</p></li>
        <li id="2"><p>Go for the Shepherd's Pie</p></li>
        <li id="3"><p>Try the Dover Sole</p></li>
        </ol>
        </blockquote>
        '''
        """
        return self.check(markup, output)

    def test_anonymous_cue(self):
        cue = ""
        line = f"<{cue}> Hello!"
        expected = textwrap.dedent(
            f"""
        <blockquote cite="&lt;&gt;">
        <p>Hello!</p>
        </blockquote>
        """
        )
        sm = SpeechMark()
        rv = sm.loads(line)
        self.compare(rv, expected, rv)

    def test_simple_cue(self):
        cue = role = "GUEST"
        line = f"<{cue}> Hello?"
        expected = textwrap.dedent(
            f"""
        <blockquote cite="&lt;{role}&gt;">
        <cite data-role="GUEST">{role}</cite>
        <p>Hello?</p>
        </blockquote>
        """
        )
        sm = SpeechMark()
        rv = sm.loads(line)
        self.compare(rv, expected, rv)

    def test_role_mode_cue(self):
        role = "GUEST"
        mode = "says"
        cue = f"{role}:{mode}"
        line = f"<{cue}> Hello?"
        expected = textwrap.dedent(
            f"""
        <blockquote cite="&lt;{cue}&gt;">
        <cite data-role="{role}" data-mode=":{mode}">{role}</cite>
        <p>Hello?</p>
        </blockquote>
        """
        )
        sm = SpeechMark()
        rv = sm.loads(line)
        self.compare(rv, expected, rv)

    def test_cue_matching_positive(self):
        examples = [
            "<>",
            "<> Hi!",
            "<role>",
            "<ROLE>",
            "<role:mode>",
            "<ROLE:MODE>",
            "<ROLE.d1.d2:mode>",
            "<ROLE:mode?p=0&q=a>",
            "<ROLE:mode#frag>",
            "<ROLE#frag>",
            "<ROLE.d1.d2:mode?p=0&q=a>",
            "<ROLE.d1.d2:mode?p=0&q=a#frag>",
        ]
        sm = SpeechMark()
        for line in examples:
            with self.subTest(line=line):
                rv = sm.cue_matcher.match(line)
                self.assertTrue(rv)
                d = rv.groupdict()
                self.assertEqual(5, len(d))

    def test_cue_matching_negative(self):
        examples = ["< >", "< >Hi!", "<role >", "< ROLE>"]
        sm = SpeechMark()
        for line in examples:
            with self.subTest(line=line):
                rv = sm.cue_matcher.match(line)
                self.assertFalse(rv)


class ParagraphTests(Syntax):
    def test_minimal_paragraph(self, markup: dict = {}, output=""):
        """
        Simple strings are encapsulated in paragraphs.

        # TOML
        markup."Plain string"  =   "Hello!"
        output = '''
        <blockquote>
        <p>Hello!</p>
        </blockquote>
        '''
        """
        return self.check(markup, output)

    def test_continuing_paragraph(self, markup: dict = {}, output=""):
        """
        A paragraph continues until explicitly terminated.

        # TOML
        markup."Continuous string"  =   '''
        Hello!
        Hello!'''
        output = '''
        <blockquote>
        <p>Hello! Hello!</p>
        </blockquote>
        '''
        """
        return self.check(markup, output)

    @Syntax.example()
    def test_multiple_paragraphs(self, markup: dict = {}, output=""):
        """
        A paragraph is terminated by a blank line.

        # TOML
        markup."Broken string"  =   '''
        Hello!

        Hello!'''
        output = '''
        <blockquote>
        <p>Hello!</p>
        <p>Hello!</p>
        </blockquote>
        '''
        """
        return self.check(markup, output)

    def test_cornercases_code(self):
        expected = textwrap.dedent(
            """
        <blockquote cite="&lt;&gt;">
        <p>Hello!</p>
        </blockquote>
        """
        )
        sm = SpeechMark()
        rv = sm.loads("<> Hello!")
        self.compare(rv, expected, rv)


class EscapingTests(Syntax):
    def test_text_escaping(self):
        sm = SpeechMark()
        for char in ">&<":
            text = f" M {char} B?"
            with self.subTest(text=text):
                rv = sm.loads(text)
                self.assertIn(html.escape(char), rv)

    def test_label_escaping(self):
        sm = SpeechMark()
        for char in ">&<":
            text = f"[M {char} B?](http://google.com)"
            with self.subTest(text=text):
                rv = sm.loads(text)
                self.assertIn(html.escape(char), rv)


class BlockTests(Syntax):
    @Syntax.example()
    def test_multiple_paragraphs(self, markup: dict = {}, output=""):
        """
        A cue is used as an attribution of speech.

        # TOML
        markup."Checking in" = '''

        <GUEST> Hello?

        Is *anyone* there?

        '''
        output = '''
        <blockquote cite="&lt;GUEST&gt;">
        <cite data-role="GUEST">GUEST</cite>
        <p>Hello?</p>
        <p>Is <em>anyone</em> there?</p>
        </blockquote>
        '''
        """
        sm = SpeechMark()
        return self.check(markup, output)

    @Syntax.example()
    def test_choice_lists(self, markup: dict = {}, output=""):
        """
        # TOML
        markup."Dialogue options" = '''
        <STAFF.suggesting#3> What would you like sir? We have some very good fish today.
            1. Order the Beef Wellington
            2. Go for the Cottage Pie
            3. Try the Dover Sole

        '''
        output = '''
        <blockquote cite="&lt;STAFF.suggesting#3&gt;">
        <cite data-role="STAFF" data-directives=".suggesting" data-fragments="#3">STAFF</cite>
        <p>What would you like sir? We have some very good fish today.</p>
        <ol>
        <li id="1"><p>Order the Beef Wellington</p></li>
        <li id="2"><p>Go for the Cottage Pie</p></li>
        <li id="3"><p>Try the Dover Sole</p></li>
        </ol>
        </blockquote>
        '''
        """
        sm = SpeechMark()
        return self.check(markup, output)


if __name__ == "__main__":
    from collections import defaultdict
    import inspect

    examples = defaultdict(list)
    for label, text, data, fn in sorted(Syntax.examples):
        info = dict(inspect.getmembers(fn))  # ["__self__"]["__class__"]
        cls = globals().get(info["__qualname__"].split(".")[0])
        examples[cls].append((label, text, data, fn))

    if __doc__:
        print(textwrap.dedent(__doc__))

    print(textwrap.dedent(Syntax.__doc__))
    for cls, entries in examples.items():
        if cls.__doc__:
            print(textwrap.dedent(cls.__doc__))
        for label, text, data, fn in entries:
            if not label:
                continue

            print(label)
            print("`" * len(label))
            print(textwrap.dedent(text))

            for note, snippet in data.get("markup", {}).items():
                print(f"{note}::\n")
                print(textwrap.indent(snippet, " " * 4))
                print()
                output = data.get("output", "")

                print("HTML5 output::\n")
                print(textwrap.indent(output, " " * 4))
                print()



#!/usr/bin/env python
#   encoding: utf-8

# Copyright (C) 2025 D E Haynes
# This file is part of spiki.

# Spiki is free software: you can redistribute it and/or modify it under the terms of the
# GNU General Public License as published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# Spiki is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with spiki.
# If not, see <https://www.gnu.org/licenses/>.

from pathlib import Path
import tempfile
import textwrap
import tomllib
import unittest

from spiki.visitor import Visitor
from spiki.plugins.indexer import Indexer
from spiki.plugins.loader import Loader


class IndexerTests(unittest.TestCase):

    def test_survey(self):
        with tempfile.TemporaryDirectory() as src_name:
            src = Path(src_name).resolve()
            index = src.joinpath("index.toml")
            index.write_text("")
            visitor = Visitor(paths=[src])
            with (
                Loader(visitor) as loader,
                Indexer(visitor) as indexer
            ):
                index_node = loader.run_enrich(path=index, node={}).node
                list(indexer.gen_survey(path=index, node=index_node))
                self.assertTrue(indexer.indexes)
                self.assertEqual(visitor.url_of(index_node), "index.html")
